// version 1.

#include <Wire.h>
#include <PWM.h>
#include <LiquidCrystal_I2C.h> 
#include <SPI.h>
#include <NRFLite.h> // Using NRFLite library
#include <EEPROM.h>

LiquidCrystal_I2C lcd(0x27,20,4); //Set  LCD I2C address

// List of subroutines
// void pw_adj()  pulse width adjustment
// void meas()  Do measurements
// void config_ads1115a  setup  ADS1115 A and start a single measurement
// void config_ads1115b  setup  ADS1115 B and start a single measurement
// int v_ads1115a() setup ADS1115 A to send measured data
// int v_ads1115b() setup ADS1115 A to send measured data
// int reg_ads1115a reading ADS1115 A measurement value
// int reg_ads1115b reading ADS1115 B measurement value
// int meas2_adc()  setup Arduino ADC to do 2 conservative measurements
// int meas4_adc()  setup Arduino ADC to do 2 conservative measurements
// void meas328()   Setup Arduino ADC to do  measurements
// void meas1115() Setup ADS1115 A & B to do measurements
// void conv_float() convert all integer measurements to floating point values
// void self_chk()  Do safety check.
// void halt(int8_t fault) On error, shut the system dowm and halt
// void in_sw()  setup the input switch
// void balance() setup batteries balancing
// void plot_results()  Print out selected measurements for serial plot
// void measure_ind()  Base on measurement, calculate the effective inductance
// void prn_results()  Print out measurements to serial monitor
// void Lcd_prn()      Output measurements to LCD 
// void cal_Ah()       calculate Amp Hour for in, out and battery
// void buttonPressed()  interrupt routine to save measured data and pause the system. 
// void err_prn(int8_t fault)  print out reason of Error
// void rf_send();   send out measured data through NRF24L01


// PWM
#define PwmBaseFreq 20000    // base PWM frequency 31.25KHz
#define PwmMaxPw    248    // max PWM pulse width
#define PwmInc      1    // PWM pulse width increment

// Limits
#define lmt_Iout  10.0    // Output current limit
#define lmt_Iin  10.0     // Input current limit
#define lmt_temp  60.0    // Temperature limit
#define lmt_intrV  0.8    //Internal input voltage limit
#define INOFF   12.2      // Input voltage threshold to shut off PWM
#define In_ON   12.0      // Input switch ON voltage threshold
#define battery_OV  12.8      // over voltage for combine batteries 
#define battery_UV  9.0      // under voltage for combine batteries 
#define out_UV  9.0      // under voltage for output

#define Max_unbal 0.2    // maximum unbalance threshold voltage 
#define Li_OV  4.3      // maximum Lithium battery OV limit
#define Li_UV  3.3      // minimum Lithium battery UD limit

#define Li_pw_thr  4.2      //  Lithium battery pulse width charging Threshold

#define ainm  2.9298/4096.0  //Arduino ADC multiple factor from Vin
#define aoutm 2.6636/4096.0  //Arduino ADC multiple factor from Vout
#define at1m 1.0086/2048.0
#define at2m 1.0082/2048.0


#define v12M 3.246 // ADC multiple factor from v12
#define v8M 2.149  // ADC multiple factor from v8
#define v4M 1.08865  // ADC multiple factor from v4

#define ILM 2.5686// ADC multiplication factor for average inductor current 
//#define IinM 2.5339// ADC multiplication factor for average inductor current 
#define IinM 2.614// ADC multiplication factor for average inductor current 
#define IoutM 2.4578 //  ADC multiplication factor for output current  
#define VintrM 4.8441  //  ADC multiplication factor for internal input voltage 
#define V5M 1.9989  //  ADC multiplication factor Vcc, 5V

// Balancing 
#define MisThr    0.01    // Balancing threshold voltage

// ADS1115 setup
#define addr1115A 0x48
#define addr1115B 0x49
#define delay1115  100    // ADS1115 conversion delay in uS
#define RemeasT  10 // re-measure time due to sudden surge of current at output
#define display_cnt 100
#define lcd_cnt_lmt 20  //define how often the LCD needed to re-initialize

// ADS1115 input selections 
uint8_t sel_A0=0b11000011;  // ADS1115A mode:A0-GND, 4.096v, single
uint8_t  sel_A1=0b11010011;  // ADS1115A mode:A1-GND, 4.096v, single
uint8_t   sel_A2=0b11100011;  // ADS1115A mode:A2-GND, 4.096v, single
uint8_t   sel_A3=0b11110011;  // ADS1115A mode:A3-GND, 4.096v, single

float ADC1b = 0.000125F;  // ADS1115 resolution at range of 4.096V.

struct measurement{
// measured by Arduino ADC
uint8_t  RadioId;
int16_t  ain;  // sum of 4 output voltages
int16_t  aout;  // sum of 4 output voltages
int16_t  at1;   // sum of 2 temperature 1
int16_t  at2;   // sum of 2 temperature 2

// Measured by ADS1115A
int16_t Iin;   //  input current
int16_t Li12;  // v12
int16_t Li8;   // v8
int16_t Li4;   // v4
// Measured by ADS1115B
int16_t IL;   // Input current

int16_t Iout;  // Output current
int16_t Vintr;   // Internal input voltage
int16_t V5;   // Vcc
// PWM

int16_t PulseWidth;
uint8_t    err;       // error code  0= no error. 
uint8_t  portb;//PORTB
uint8_t  portd; //PORTD  value
uint16_t SendCounts;  // number packages sent
};

// measured by Arduino ADC
float  faout;  // output voltages
float  fat1;   // temperature 1
float  fat2;   // temperature 2

float pwr_in;  // input power
float pwr_out;  // output power
float sum_pwr_in; // sum of 100 input power
float avg_pwr_in,avg_prv_pwr; // average input power and prv_pwr
float    pwr_bat; // battery power  pos= charging  neg= discharging
float    i_bat;  // current flows into battery, pos= charging  neg= discharging

// Measured by ADS1115A
float fVin; // solar input voltage
float fLi12;  // v12
float fLi8;   // v8
float fLi4;   // v4
// Measured by ADS1115B
float fIL; // inductor  current
float fIin;  // input current
float fIout;  // output current
float fVintr;   // internal input voltage
float fV5;   // Vcc

float vb12,vb8,vb4;  //  voltage across battery
float vbmis;   //voltage difference between battery

// int16_t junk=0;  // temperatory storage
long freq = PwmBaseFreq;

float pwr_on_t;  // Total power on time in sec.

//******************************************************************

uint8_t meas2EEPROM;  //Press button Interrupt routine will set this to 1 inorder to save measurements into EEPROM
//******************************************************************

// Hardware configuration: Set up nRF24L01 radio on SPI bus (pins  11, 12, 13) plus pins 8 & 10
#define RF_CHAN 108  //Using Radio Channel
#define RADIO_TX_ID  13             // Transmitter Radio ID. Not use in this program
#define DESTINATION_RADIO_ID 18 // Our Receiving Radio ID. 
#define PIN_RADIO_CE  8
#define PIN_RADIO_CSN 10

NRFLite _radio;
measurement vm;
//******************************************************************
int lcd_cnt;  // counts LCD display update
//******************************************************************

void setup()
{
    Serial.begin(115200);

    // By default, 'init' configures the radio to use a 2MBPS bitrate on channel 100 (channels 0-125 are valid).
    // Both the RX and TX radios must have the same bitrate and channel to communicate with each other.
    // You can run the 'ChannelScanner' example to help select the best channel for your environment.
    // You can assign a different bitrate and channel as shown below.
    //   _radio.init(RADIO_ID, PIN_RADIO_CE, PIN_RADIO_CSN, NRFLite::BITRATE2MBPS, 100) // THE DEFAULT
    //   _radio.init(RADIO_ID, PIN_RADIO_CE, PIN_RADIO_CSN, NRFLite::BITRATE1MBPS, 75)
    //   _radio.init(RADIO_ID, PIN_RADIO_CE, PIN_RADIO_CSN, NRFLite::BITRATE250KBPS, 0)
    
    if (!_radio.init(DESTINATION_RADIO_ID, PIN_RADIO_CE, PIN_RADIO_CSN, NRFLite::BITRATE250KBPS, RF_CHAN))
    {
        Serial.println(" Initial communicate with radio fail");
        while (1); // Wait here forever.
    }
    else
       Serial.println(" NRF24L01 Connected");
  // Set I2C bus as a master 
  Wire.begin();  // initialize I2C
  scan_i2c();  // scan for I2C devices
       
// Initialize LCD display
  lcd.init();   //initialize LCD
  lcd.backlight();  // turn LCD light on  
  lcd.begin(20, 4); // set up the LCD's number of columns and rows:
  lcd.clear(); // clear LCD screen

}

void loop()
{
    while (_radio.hasData()) {
      _radio.readData(&vm); // Reading RF data.
      conv_float();  // converted to floating point measurements
      prn_results();    //print vm data
      Lcd_prn();  //Display measurement at LCD
   }
}
//******************************************************************
//***********************************************************************
void conv_float(){  

  // convert all measurement to floating point
    fLi12 =v12M *ADC1b * vm.Li12;  // voltage at v12 reference to GND 
    fLi8 = v8M *ADC1b * vm.Li8;  // voltage at v8 reference to GND
    fLi4 = v4M *ADC1b * vm.Li4;  // voltage at v4 reference to GND
    fIL = ILM *ADC1b * vm.IL;  // Inductor  current
    fIin = IinM *ADC1b * vm.Iin;  // Input  current
   if (fIL < 0.0) fIL = 0.0;
    fIout = IoutM *ADC1b * vm.Iout;  // output current
    if (fIout < 0.0) fIout = 0.0;
    fVintr = VintrM *ADC1b * vm.Vintr;  // Input voltage internal 
    fV5 = V5M *ADC1b * vm.V5;    // Vcc=5V
//    fV5 = 4.9958;   // V5 fix due to fail in ADS1115, measured by DVM
    faout = aoutm*fV5*vm.aout;     // Output voltage
    fVin = ainm*fV5*vm.ain;     // Input voltage
  

    fat1 = at1m*fV5*vm.at1;        // temperature voltage at probe 1
    //convert voltage to temperature, equation is base on measurement.+/- 5C between 20C to 80C 
    if (fat1 < 3.455)  fat1 = 25.5*fat1 -17.5;
    else if (fat1 < 4.095) fat1 = 39.2*fat1 -64.1;
         else fat1 = 75.8*fat1 -215.2;

    fat2 = at2m*fV5*vm.at2;        // temperature voltage at probe 2
    //convert voltage to temperature, equation is base on measurement.+/- 5C between 20C to 80C 
    if (fat2 < 3.483)  fat2 = 25.2*fat2 -17.2;
    else if (fat2 < 4.143) fat2 = 37.9*fat2 -60.9;
         else fat2 = 921*fat2 -286.3; 
 
    vb12 = fLi12 - fLi8;  // v12 Li battery voltage
    vb8 = fLi8 - fLi4;    // v8 Li battery voltage
    vb4 = fLi4;           // v4 Li battery voltage

    vbmis = max(vb12, max(vb8,vb4)) - min(vb12, min(vb8,vb4));  //max difference in Li battery voltages 
/*    ind_dv = fIL-fLi12; // voltage across inductor
    i_bat = fIL -fIout;
    
    pwr_in = fVin*fIin;
    pwr_out = fLi12 * fIL;
    sum_pwr_in+=pwr_in;*/
}
//***********************************************************************
//  print measurements
void prn_results() {

   Serial.print("Data from Radio ");  
   Serial.print(vm.RadioId);
   Serial.print("  Sent # ");
   Serial.print(vm.SendCounts);   
   Serial.print(F("\t\tERROR "));
   Serial.println(vm.err);
   
   Serial.print(F("Vin, V\t"));  
   Serial.print(F("\tV12, V\t"));
   Serial.print(F("\tV8, V\t"));
   Serial.print(F("\tV4, V\t")); 
   Serial.print(F("\tVout,V"));
   Serial.println();      
   Serial.print(fVin,3);
   Serial.print(F("\t\t"));  
   Serial.print(fLi12,3);
   Serial.print(F("\t\t"));  
   Serial.print(fLi8,3);
   Serial.print(F("\t\t"));  
   Serial.print(fLi4,3);
   Serial.print(F("\t\t"));  
   Serial.print(faout,3);
   Serial.println();
   
   Serial.print(F("Vintr, V\t"));
   Serial.print(F("Vb12, V\t"));  
   Serial.print(F("\tVb8, V\t"));
   Serial.print(F("\tVb4, V\t"));
   Serial.print(F("\tvbmis, V"));
   Serial.println();
   Serial.print(fVintr,3);
   Serial.print(F("\t\t"));        
   Serial.print(vb12,3);
   Serial.print(F("\t\t"));  
   Serial.print(vb8,3);
   Serial.print(F("\t\t"));  
   Serial.print(vb4,3);
   Serial.print(F("\t\t"));  
   Serial.print(vbmis,3);
   Serial.println();
              
   Serial.print(F("Iin, A\t"));
   Serial.print(F("\tIL, A\t"));
   Serial.print(F("\tIout, A\t"));  
   Serial.print(F("\tI_bat, A"));
   Serial.print(F("\tV5, V"));
   Serial.println();

   Serial.print(fIin,3);
   Serial.print(F("\t\t"));  
   Serial.print(fIL,3);
   Serial.print(F("\t\t"));  
   Serial.print(fIout,3);
   Serial.print(F("\t\t")); 
   Serial.print(i_bat ,3);
   Serial.print(F("\t\t"));
   Serial.print(fV5,3);
   Serial.println();
             
   Serial.print(F("Freq Hz\t"));  
   Serial.print(F("\tPulse Width"));   
   Serial.print(F("\tTemp1, C"));  
   Serial.print(F("\tTemp2, C"));   
   Serial.print(F("\tPORTD"));  
   Serial.println(); 
   
   Serial.print(freq);
   Serial.print(F("\t\t"));  
   Serial.print(vm.PulseWidth);
   Serial.print(F("\t\t"));  
   Serial.print(fat1,1);
   Serial.print(F("\t\t"));  
   Serial.print(fat2,1);
   Serial.print(F("\t\t"));  
   Serial.print(PORTD,BIN);  
   Serial.println();  
   Serial.println();  
}


//**********************************************************
// Print measurements to LCD display
void Lcd_prn() {
  uint8_t portd;
  static int a;

// LCD display VChg, IChg, IOut, TC  
// re-initialize at every 30sec or 20 lcd counts so the LCD can swap out
 
  if ( lcd_cnt > lcd_cnt_lmt) {
  lcd.init();   //initialize LCD
  lcd.backlight();  // turn LCD light on  
  lcd.begin(20, 4); // set up the LCD's number of columns and rows:    lcd_cnt =0;
  lcd_cnt = 0; 
  }
 else lcd_cnt++;


  lcd.clear();
  lcd.setCursor(0, 0);  // set the cursor to column 0, row 1
// 1st row, v12,v8,v4, * indicate balancing on that battery
  if((vm.portd & 0b0010000)==0);
  else  lcd.print(F("*"));
  lcd.print(fLi12,3);  // v12  voltage, 6 chars
  if((vm.portd & 0b00001000)==0)    lcd.print(F(" "));
  else  lcd.print(F(" *"));
  lcd.print(fLi8,3);  // v8  voltage, 5 chars 
  if((vm.portd & 0b10000000)==0)    lcd.print(F(" *"));
  else  lcd.print(F(" "));
  lcd.print(fLi4,3);  // v4  voltage, 5 chars

  // row 1,  Vin, Vout, pulse width, PortD
  lcd.setCursor(0, 1);  // set the cursor to column 0, row 1
  lcd.print(fVin,2);  // vin  voltage
  lcd.print(F(" "));
  lcd.print(fVintr,2);  // Internal vin 
  lcd.print(F(" "));
  lcd.print(faout,2);  // vout  voltage
  lcd.print(F(" "));
  lcd.print(vm.portd,HEX);  //  PORTD, 2 chars
  
  // row 2,  Vin, Vout, pulse width, PortD 
  lcd.setCursor(0, 2);  // set the cursor to column 0, row 1 
  lcd.print(fIL,1);   //charging current,5 chars
  lcd.print(F("A "));  // 
  lcd.print(fIout,1);   //output current, 5 chars
  lcd.print(F("A "));
  lcd.print(vm.PulseWidth);  // pulse width
  lcd.print(F(" T"));
  lcd.print(fat1,0);  //Temperature 1, 4 chars
  lcd.print(F(" "));
  lcd.print(fat2,0);  //Temperature 1, 5 chars    

  lcd.setCursor(0, 3);  // set the cursor to column 0, row 3 
  if (vm.err > 0) {
  lcd.print(F("E"));
  lcd.print(vm.err);  //Error code, 1 chars   
  lcd.print(F(" B 0x"));
  lcd.print(vm.portb,HEX);  //Error code, 1 chars     
  lcd.print(F(" D 0x"));
  lcd.print(vm.portd);  //Error code, 1 chars   
  }
  else {
  lcd.print(vb12,3);   //charging current,5 chars
  lcd.print(F(" "));  // 
  lcd.print(vb8,3);   //output current, 5 chars
  lcd.print(F(" "));
  lcd.print(1000*vbmis,0);  // Mismatch  voltage, 5 chars
  lcd.print(F("m E"));  
  lcd.print(vm.err,BIN);  //Temperature 1, 4 chars 
  }
}
 
//***********************************************************
void scan_i2c(){
 Serial.println("Scanning I2C");

  for (byte address = 1; address < 127; ++address) {
    // The i2c_scanner uses the return value of
    // the Write.endTransmisstion to see if
    // a device did acknowledge to the address.
    Wire.beginTransmission(address);
    byte error = Wire.endTransmission();

    if (error == 0) {
      Serial.print(F("Found I2C Addr 0x"));
      Serial.println(address, HEX);
      }
  }
}


//**********************************************************
